/*
* TC18 by Jason von Nieda <jason@vonnieda.org>
* License: CC BY-SA
* 
* Board Configuration
*  Arduino Duemilanove or Diecimila, ATmega328P or ATmega168 @ 16MHz
*  Arduino will set the brown out detect too high on burning the bootloader
*  so it has to be changed to 1.8 manually.
*  VBB feedback       Analog 0 (PC0) // ice: not used
*    VBB with 100k to 1k voltage divider
*  SCK to MAX6921     Digital 13 (PB5)
*  MOSI to MAX6921    Digital 11 (PB3)
*  PPS                Digital 8 (PB0) // ice: not used
*  SW1 (ENC_A)        Digital 6 (PD6) // ice: PD5
*  SW2 (ENC_B)        Digital 7 (PD7) // ice: PB0
*  SW3 (ENC_SW)       Digital 5 (PD5) // ice: PD4
*  LOAD               Digital 4 (PD4) // ice: PC0
*  BLANK                              // ice: PC3
*  BOOST              Digital 3 (PD3) // icetube: PD6 
*  SDA                Analog 4 (PC4)
*  SCL                Analog 5 (PC5)
*  RXD                Digital 0 (PD0)
*  TXD                Digital 1 (PD1)
*/
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdio.h>
#include <Wire.h>
#include <EEPROM.h>
#include "data.h"

volatile byte vfd_data[9];
volatile byte vfd_digit;
// Digit dwell is used to adjust the brightness of individual digits on
// the display. Used to compensate for tube manufacturing differences.
// Increasing the value of a position causes the refresh loop to dwell on
// that digit for additional cycles causing it's brightness to increase.
// Use sparingly. This slows the refresh of the entire display.
// This feature is not yet complete. Needs to be moved to a configuration
// menu and stored in config.
byte vfd_digit_dwell[] = {
  0, 0, 0, 0, 0, 0, 0, 0, 1
};
volatile byte vfd_digit_dwell_count;
volatile unsigned long spi_buffer;
volatile byte spi_buffer_count;

// Default boost voltage used during startup, this is quickly replaced by the
// config data
//float voltage_target = 30.00;
// The allowable voltage +/- drift before the ADC interrupt makes adjustements
//static const float voltage_drift = 3.00;
// The current voltage as measured by the ADC
//float voltage;
// The number of times the voltage has been sampled. This is used to limit
// the frequency of updates to the boost PWM.
//byte voltage_sample;

volatile char encoder_debounce = -1;
volatile char encoder_pos;
volatile char encoder_pressed;

volatile byte hour, minute, second, year, month, date;

volatile unsigned long millisecond1 = 0;
volatile float temperature;

// Used in many main loop functions to format text for display
char display_buf[20];

// Configurable options, pulled from EEPROM at startup and stored
// when changed.
char display_mode;
char brightness;
char bank_mode;

// used because we are not using RTC's 1s SQW
volatile unsigned int millisecond;

byte serial_mode = 0;
// **********************************************
//                     SETUP
// **********************************************

void setup() {
  Serial.begin(9600);
  Serial.println("welcome");
  // ground blank pin
  pinMode(17, OUTPUT);
  digitalWrite(17, LOW);
  
  BOOST_DDR |= _BV(BOOST);

  // Timer 2 is used to refresh the VFD display, 1ms interupts, 
  // Set defaults in case Arduino screws them up
  TCCR2A = TCCR2B = TIMSK2 = 0;  
  TCCR2A = 0;
  // Clock Select, prescaler = 64
  TCCR2B = _BV(CS22);
  // Initial timer count
  TCNT2 = 0xC2;
  // Set overflow interupt
  TIMSK2 = _BV(TOIE2);

// I understand timer0 is used for system timing on arduino
  // fast PWM, set OC0A (boost output pin) on match
  TCCR0A = TCCR0B = TIMSK0 = 0;
  TCCR0A = _BV(WGM00) | _BV(WGM01);  
  // Use the fastest clock
  TCCR0B = _BV(CS00);
  TCCR0A |= _BV(COM0A1);
  TIMSK0 |= _BV(TOIE0);
  OCR0A  = 30;

  TCCR1A = TCCR1B = TIMSK1 = 0;

  // Enable SS, MOSI and SCK as outputs, required by SPI
  DDRB |= _BV(PORTB2) | _BV(PORTB3) | _BV(PORTB5);
  // SPI Enable, Master Mode, Interrupt Enable, FOSC/16
  SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR0) | _BV(SPIE);
 
  VFDLOAD_DDR |= _BV(VFDLOAD);

// Startup, beep
  ds3231_setup();
  load_config();
  speaker_init();
  beep(4000, 1);
  
// Turn on VFD
  pinMode(3, OUTPUT);
  digitalWrite(3, LOW);
  
  // Set ENC_A, ENC_B and ENC_SW pins to input
  ENC_A_DDR &= ~(_BV(ENC_A));
  ENC_B_DDR &= ~(_BV(ENC_B));
  ENC_SW_DDR &= ~(_BV(ENC_SW));
  
  // Turn on internal pullups for ENC_A, ENC_B, ENC_SW
  ENC_A_PORT |= _BV(ENC_A);
  ENC_B_PORT |= _BV(ENC_B);
  ENC_SW_PORT |= _BV(ENC_SW);

  pinMode(ALARM_ARDUINO, INPUT);
  digitalWrite(ALARM_ARDUINO, HIGH);
  
  // Enable Pin Change Interrupt 2 for PCINT22 (PD6) [ENC_A], PCINT21 (PD5) [ENC_SW]
  //PCMSK2 |= _BV(PCINT21);
  //PCMSK2 |= _BV(PCINT22);
  PCMSK0 = _BV(PCINT0); // Button 2
  PCMSK2 |= _BV(PCINT20); // PD4
  PCMSK2 |= _BV(PCINT21); // PD5
  PCICR |= _BV(PCIE0) | _BV(PCIE2);

  EICRA = _BV(ISC00);
  EIMSK = _BV(INT0);


  if (digitalRead(ALARM_ARDUINO)) {
    serial_mode = 1;
    sprintf(display_buf, "Ser On");
    vfd_set_string(display_buf, 0);
    vfd_data[0] = 0;
    delay2(2000);
    clear_display();
  } 

}



int serial_readline(char *s, int length) {
  int ch;
  char i = 0;
  do {
    ch = Serial.read();
    *s++ = ch;
    i ++;
    delay2(5);
  } while (ch != '\n' && ch != '\r' && Serial.available() > 0);
  *s = NULL;
  return i;
}



/*
  Timer fires approx. every 1ms and handles refreshing the VFD,
  debouncing inputs and updating the milliscond time register.
*/
ISR(TIMER2_OVF_vect) {
  // fire approx every 1ms
  TCNT2 = 0xC2;
  millisecond++;
  millisecond1++;
  // Debug
 /* if (millisecond1 % 50 == 0) {
    Serial.println (millisecond1);
  }*/
  
  if (encoder_debounce >= 0) {
    encoder_debounce--;
  }
  
  if (encoder_debounce == 0) {
    int encoder_a = (ENC_A_PIN & _BV(ENC_A));
    int encoder_b = (ENC_B_PIN & _BV(ENC_B));

    int encoder_sw = (ENC_SW_PIN & _BV(ENC_SW));
    if (!encoder_sw) {
      encoder_pressed++;
    }
    
    if (!encoder_a) {
      encoder_pos = -1; 
    } else if (!encoder_b) {
      encoder_pos = 1; 
    } else {
      encoder_pos = 0;
    }
    
    PCIFR |= _BV(PCIF2);
    PCICR |= _BV(PCIE2);
    PCIFR |= _BV(PCIF0);
    PCICR |= _BV(PCIE0);    
  }
  
  vfd_refresh();
  // update RTC every 500ms
 if (millisecond >= 500) {
    millisecond = 0;
    updateRTC();
 }
  
}

// Because I broke delay by messing with timer0
void delay2(unsigned long ms)
{
  unsigned long start = millisecond1;
  while (millisecond1 - start <= ms) { ;}
		
}


/*
  Handles the completion of a SPI transfer and either latches
  the data or starts the transfer of the next byte
*/
ISR(SPI_STC_vect) {
  if (spi_buffer_count) {
    SPDR = (spi_buffer >> 24);
    spi_buffer <<= 8;
    spi_buffer_count--;
  }
  else {
    // latch data
    VFDLOAD_PORT |= _BV(VFDLOAD);
    VFDLOAD_PORT &= ~_BV(VFDLOAD);
  }
}



void load_config() {
  byte b;

  b = EEPROM.read(0);
  if (b == 255) {
    b = 0;
  }
  display_mode = b;
  
  b = EEPROM.read(1);
  if (b == 255) {
    b = 2;
  }
  brightness = b;
  brightness_to_timer();
  
  b = EEPROM.read(2);
  if (b == 255) {
    b = 1;
  }
  bank_mode = b;
  
}

void save_config() {
  EEPROM.write(0, display_mode);
  EEPROM.write(1, brightness);
  EEPROM.write(2, bank_mode);
}



float c_to_f(float c) {
  return (((9.0 / 5.0) * c) + 32.0);
}

/*
  Wait the given number of milliseconds (ms) for any input from the encoder.
  If there is input the enc_moved and enc_pressed references are filled in
  and the function returns true. If no input is detected before the
  timer expires the function returns false.
  This function clears the global input counters in either case.
*/
boolean wait_for_input(int ms, char *enc_pos, boolean *enc_pressed) {
  unsigned long m = millisecond1 + ms;
  while (!encoder_pressed && !encoder_pos && (millisecond1 < m));
  *enc_pos = encoder_pos;
  *enc_pressed = encoder_pressed;
  encoder_pressed = 0;
  encoder_pos = 0;
  return *enc_pressed || *enc_pos;
  return false;
}

/*
  The main loop of the program is responsible for displaying the current time, checking for input
  from the encoder or button and running the menu system for the user to interact with the clock.
*/
void loop() {
  char enc_moved = 0;
  boolean enc_pressed = 0;
  char scroll_on = 0;
  for (;;) {
    if (serial_mode == 0) {
      // This effectively defines the refresh rate for the clock data. We wait for input
      // for 10 ms and then refresh the clock.
      if (wait_for_input(500, &enc_moved, &enc_pressed)) {
        main_menu();
      }
      switch (display_mode) {
        case 0: {
          sprintf(display_buf, "%02d:%02d:%02d", hour, minute, second);
          vfd_data[0] = 0;
          break;
        }
        case 1: {
          sprintf(display_buf, "%02d%02d  %02d", hour, minute, date);
          vfd_data[0] = (second % 2 ? VFD_SEG_H : 0);
          break;
        }
        case 2: {
          sprintf(display_buf, "%02d%02d %dF", hour, minute, (int) c_to_f(temperature));
          vfd_data[0] = (second % 2 ? VFD_SEG_H : 0);
          break;
        }
        case 3: {
          sprintf(display_buf, "%s %02d", month_abbrevs[month - 1], date);
          vfd_data[0] = 0;
          break;
        }
        case 4: {
          uint64_t t = (24L * 60L * 60L * 1000L) - ((hour * 60L * 60L * 1000L) + (minute * 60L * 1000L) + (second * 1000L) + millisecond);
          sprintf(display_buf, "%08lu", t);
          vfd_data[0] = 0;
          break;
        }
        case 5: {
          sprintf(display_buf, "%02d%02d  %02d", hour, minute, (int) c_to_f(temperature));
          vfd_data[0] = (second % 2 ? VFD_SEG_H : 0);
          break;
        }
      }
      if (bank_mode) {
        if (second % 12 < 4) {
            sprintf(display_buf, "%02d-%02d-%02d", month, date, year);
            vfd_data[0] = 0;
        }
        else if (second % 12 < 8) {
            sprintf(display_buf, "%dF %dC", (int) c_to_f(temperature), (int) temperature);
            vfd_data[0] = 0;
        }
        
      }
      //Serial.println("mew");
      if (serial_mode == 0) {
        vfd_set_string(display_buf, 0);
      }
    } else {
      if (wait_for_input(100, &enc_moved, &enc_pressed)) {
        main_menu();
      }
      if (Serial.available() > 0) {
        unsigned char i = serial_readline(display_buf, 20);
        vfd_data[0] = 0;
        if (i > 8) {
           vfd_scroll_string(display_buf, 200);
           scroll_on = 1; // set to scroll
        } else {
           vfd_set_string(display_buf, 0);
           scroll_on = 0; // stop scroll
        } 
      }
      // Keep scrolling if we need to
      if (scroll_on) {
           vfd_scroll_string(display_buf, 200);
      }      
     
    }
  }
}



